package ru.ifmo.vizi.ukkonen.SuffixTree;

import ru.ifmo.vizi.base.widgets.Shape;
import ru.ifmo.vizi.base.widgets.ShapeStyle;

import java.awt.*;

/**
 * Edge class.
 *
 * @author  Igor Ahmetov
 * @version $Id$
 */

public class Edge extends Shape {
    static public double arcHeight = 0.15;

    /**
     * Arrow height.
     */
    static public double arrowHeight = 10;

    /**
     * Arrow width.
     */
    static public double arrowWidth = 4;

    /**
     * Beginning of the edge.
     */
    protected NodeShape from;

    /**
     * End of the edge.
     */
    protected NodeShape to;

    /**
     * Stores <code>true</code> if this edge has acquired the edge lock.
     */
    protected boolean hasLock = false;

    protected boolean hasArrow;

    protected boolean hasReverseArrow;

    /**
     * Default constructor.
     *
     * @param styleSet edge style set.
     * @param from beginning of the edge.
     * @param to end of the edge.
     */
    public Edge(ShapeStyle[] styleSet, NodeShape from, NodeShape to) {
        super(styleSet);

        this.from = from;
        this.to = to;
    }

    /**
     * Returns source node of this edge.
     *
     * @return source node of this edge.
     */
    public NodeShape getSourceNodeShape() {
        return from;
    }

    /**
     * Returns target node of this edge.
     *
     * @return target node of this edge.
     */
    public NodeShape getTargetNodeShape() {
        return to;
    }

    /**
     * Checks if the edge contains the specified point.
     *
     * @param x x coordinate of the point.
     * @param y y coordinate of the point.
     * @return <code>true</code> if the edge contains the
     *         specified point.
     */
    public boolean containsPoint(int x, int y) {
        Rectangle fromBounds = from.getBounds();
        Rectangle toBounds = to.getBounds();
        int x1 = fromBounds.x + fromBounds.width / 2;
        int y1 = fromBounds.y + fromBounds.height / 2;
        int x2 = toBounds.x + toBounds.width / 2;
        int y2 = toBounds.y + toBounds.height / 2;

        int a = y1 - y2, b = x2 - x1, c = -a * x1 - b * y1;
        double d = Math.abs(a * x + b * y + c) / Math.sqrt(a * a + b * b);

        return (d <= 8 && x >= Math.min(x1, x2) - d && x <= Math.max(x1, x2) + d
                && y >= Math.min(y1, y2) - d && y <= Math.max(y1, y2) + d);
    }

    /**
     * Checks if this <code>Edge</code> object has the edge lock.
     *
     * @return <code>true</code> if this edge has the lock.
     */
    public boolean hasLock() {
        return hasLock;
    }

    /**
     * Acquires the edge lock.
     *
     * @return <code>true</code> if the lock has been acquired.
     */
    public boolean acquireLock() {
        if (hasLock) return false;

        SuffixTree tree = (SuffixTree) getParent();
        if (tree == null) return false;
        hasLock = tree.acquireLock(this);
        return hasLock;
    }

    /**
     * Releases the edge lock.
     *
     * @return <code>true</code> if the lock has been released.
     */
    public boolean releaseLock() {
        if (!hasLock) return false;

        SuffixTree tree = (SuffixTree) getParent();
        if (tree == null) return false;
        tree.releaseLock(this);
        hasLock = false;
        return true;
    }

    protected Dimension fit(Dimension size) {
        return size;
    }

    public void setArrow(boolean hasArrow) {
    	this.hasArrow =  hasArrow;
    }

    public void setReverseArrow(boolean hasReverseArrow) {
    	this.hasReverseArrow = hasReverseArrow;
    }

    protected double getAngle(double x, double y) {
        double a = (x == 0) ? (y >= 0 ? Math.PI / 2 : Math.PI * 3 / 2) :
                Math.atan(y / x);
        if (x < 0) a += Math.PI;
        return a;
    }


    /**
     * Paints this edge.
     *
     * @param g graphics context for painting.
     */
    public void paint(Graphics g) {
        if (look == null || g == null || to == null || from == null) return;

        Rectangle fromBounds = from.getBounds();
        if (fromBounds == null) return;

        Rectangle toBounds = to.getBounds();
        if (toBounds == null) return;

        int x1 = fromBounds.x + fromBounds.width / 2;
        int y1 = fromBounds.y + fromBounds.height / 2;
        int x2 = toBounds.x + toBounds.width / 2;
        int y2 = toBounds.y + toBounds.height / 2;

        if (look.getBorderStatus(style)) {
            g.setColor(look.getBorderColor(style));
            g.drawLine(x1, y1, x2, y2);
        }

        if (hasArrow) {
            double angle = getAngle(x2 - x1, y2 - y1) + Math.PI;
            double r = Math.min(toBounds.width / 2, toBounds.height / 2);

            int x[] = new int[3], y[] = new int[3];
            x[0] = x2 + (int) Math.round(r * Math.cos(angle));
            y[0] = y2 + (int) Math.round(r * Math.sin(angle));
            x[1] = x[0] + (int) Math.round(arrowHeight * Math.cos(angle) + arrowWidth * Math.sin(angle));
            y[1] = y[0] + (int) Math.round(arrowHeight * Math.sin(angle) - arrowWidth * Math.cos(angle));
            x[2] = x[0] + (int) Math.round(arrowHeight * Math.cos(angle) - arrowWidth * Math.sin(angle));
            y[2] = y[0] + (int) Math.round(arrowHeight * Math.sin(angle) + arrowWidth * Math.cos(angle));

            if (look.getFillStatus(style)) {
                g.setColor(look.getFillColor(style));
                g.fillPolygon(x, y, 3);
            }
            if (look.getBorderStatus(style)) {
                g.setColor(look.getBorderColor(style));
                g.drawPolygon(x, y, 3);
            }
        }

        if (hasReverseArrow) {
            double angle = getAngle(x1 - x2, y1 - y2) + Math.PI;
            double r = Math.min(fromBounds.width / 2, fromBounds.height / 2);

            int x[] = new int[3], y[] = new int[3];
            x[0] = x1 + (int) Math.round(r * Math.cos(angle));
            y[0] = y1 + (int) Math.round(r * Math.sin(angle));
            x[1] = x[0] + (int) Math.round(arrowHeight * Math.cos(angle) + arrowWidth * Math.sin(angle));
            y[1] = y[0] + (int) Math.round(arrowHeight * Math.sin(angle) - arrowWidth * Math.cos(angle));
            x[2] = x[0] + (int) Math.round(arrowHeight * Math.cos(angle) - arrowWidth * Math.sin(angle));
            y[2] = y[0] + (int) Math.round(arrowHeight * Math.sin(angle) + arrowWidth * Math.cos(angle));

            if (look.getFillStatus(style)) {
                g.setColor(look.getFillColor(style));
                g.fillPolygon(x, y, 3);
            }
            if (look.getBorderStatus(style)) {
                g.setColor(look.getBorderColor(style));
                g.drawPolygon(x, y, 3);
            }
        }

        if (to.getNode() == null) return;
        String str = to.getNode().getString();
        if (str == null) return;

        int edgeWidth = Math.abs(x1 - x2);
        int edgeHeight = Math.abs(y1 - y2);

        int fontSize = Math.min(30, Math.max(edgeHeight / (str.length() + 3),
                                    edgeWidth / 18));

        if (fontSize < 12 && to.getNode().getString().length() > 7) {
        	String nodeString = to.getNode().getString();
        	int len = nodeString.length();

            str = "" + nodeString.charAt(0) + nodeString.charAt(1) + ".." +
            		nodeString.charAt(len - 2) + nodeString.charAt(len - 1);

            fontSize = Math.min(30, Math.max(edgeHeight / (str.length() + 3),
                                    edgeWidth / 18));            		            		
        }

        if (fontSize < 8) return;

        if (getFont() == null) return;
        Font font = new Font(getFont().getName(), getFont().getStyle(), fontSize);
        if (font == null) return;
        g.setFont(font);

        for (int i = 0; i < str.length(); i++) {
            double newx, newy;
            double k = 0.025;
            double x = x1 + (x2 - x1) * (i + 2) / (str.length() + 3);
            double y = y1 + (y2 - y1) * (i + 2) / (str.length() + 3);

            int a = x2 - x1, b = y2 - y1;
     
            Container parent = getParent();
            if (parent == null || parent.getSize() == null) return;
            int size = Math.min(parent.getSize().width,
            		parent.getSize().height);

            FontMetrics fm = g.getFontMetrics();
            if (fm == null) return;
            int charHeight = fm.getAscent();
            int charWidth = fm.charWidth(str.charAt(i));

            if (x2 >= x1) {
            	newx = x + k * b / Math.sqrt(a * a + b * b) * size;
                newy = y - k * a / Math.sqrt(a * a + b * b) * size;
                g.drawString(str.substring(i, i + 1),
                		(int) Math.round(newx), (int)Math.round(newy + (double) charHeight / 3));
            } else {
                newx = x - k * b / Math.sqrt(a * a + b * b) * size;
                newy = y + k * a / Math.sqrt(a * a + b * b) * size;
                g.drawString(str.substring(i, i + 1),
                		(int) Math.round(newx - charWidth),
                        (int) Math.round(newy + (double) charHeight / 3));
            }
        }
    }

    /**
     * Paints hint of this edge on specified <code>Graphics</code> object.
     *
     * @param g graphics context for painting.
     */
    public void paintHint(Graphics g) {
        if (g == null) return;
        if (hasLock) {
            if (getSize() == null || getParent() == null || getFont() == null) return;
            Font font = new Font(getFont().getName(), getFont().getStyle(),
                                (int) Math.round(0.08 * getParent().getSize().height));
            if (font == null) return;
            g.setFont(font);

            if (to.getNode() == null) return;
            String str = to.getNode().getString();

            FontMetrics fm = g.getFontMetrics();
            if (fm == null) return;
            int charHeight = fm.getAscent();
            int stringWidth = fm.stringWidth(str);

            Point mousePoint = ((SuffixTree) getParent()).getMousePoint();
            if (mousePoint == null) return;

            int x = mousePoint.x, y = mousePoint.y;
            int width = stringWidth + 10, height = charHeight + 5;

            if (y - height < 0) {
                x -= width;
                y += height;
            } else
                if (x + width > getSize().width)
                    x -= width;

            g.clearRect(x, y - height, width, height);
            g.drawLine(x, y, x, y - height);
            g.drawLine(x, y - height, x + width, y - height);
            g.drawLine(x + width, y - height, x + width, y);
            g.drawLine(x + width, y, x, y);

            g.drawString(str, x + 5, y - 5);
        }
    }
}