package ru.ifmo.vizi.ukkonen.SuffixTree;

import ru.ifmo.vizi.base.widgets.Ellipse;
import ru.ifmo.vizi.base.widgets.ShapeStyle;

import java.awt.*;

/**
 * Component that represents suffix tree node.
 *
 * @author	Igor Ahmetov
 * @version	$Id$
 */
public class NodeShape extends Ellipse {
    /**
     * Parent container.
     */
    protected SuffixTree tree;

    /**
     * Corresponding node in the suffix tree.
     */
    protected SuffixTreeNode node;

    /**
     * Depth of this node in the suffix tree.
     */
    protected int depth;

    /**
     * Horizontal position of this node in the suffix tree.
     */
    protected int hpos;

    /**
     * Constructs new <code>NodeShape</code> object.
     *
     * @param styleSet node style set.
     * @param tree parent container.
     * @param node corresponding node in the suffix tree.
     */
    public NodeShape(ShapeStyle[] styleSet, SuffixTree tree, SuffixTreeNode node) {
        this(styleSet, tree, node, "");
    }

    /**
     * Constructs new <code>NodeShape</code> object with
     * specified message on it.
     *
     * @param styleSet node style set.
     * @param tree parent container.
     * @param node corresponding node in the suffix tree.
     * @param message message on the node.
     */
    public NodeShape(ShapeStyle[] styleSet, SuffixTree tree, SuffixTreeNode node,
                     String message) {
        super(styleSet, message);

        this.tree = tree;
        this.node = node;
    }

    /**
     * Returns component that represents this nodes parent in the suffix tree.
     *
     * @return component.
     */
    public NodeShape getParentShape() {
        return tree.getNodeShape(node.getParent());
    }

    /**
     * Returns component that represents this nodes leftmost child in the suffix tree.
     *
     * @return component.
     */
    public NodeShape getLeftChildShape() {
        return tree.getNodeShape(node.getLeftChild());
    }

    /**
     * Returns component that represents this nodes right sibling in the suffix tree.
     *
     * @return component.
     */
    public NodeShape getRightSiblingShape() {
        return tree.getNodeShape(node.getRightSibling());
    }

    /**
     * Returns tree node associated with this shape.
     *
     * @return tree node.
     */
    public SuffixTreeNode getNode() {
        return node;
    }

    /**
     * Returns depth in the suffix tree of this node.
     *
     * @return depth in the suffix tree.
     */
    public int getDepth() {
        return depth;
    }

    /**
     * Returns horizontal position of this node.
     *
     * @return horizontal position of this node.
     */
    public int getHorizontalPos() {
        return hpos;
    }

    /**
     * Checks if the node contains the specified point.
     *
     * @param point point to check.
     * @return <code>true</code>, if node contains the specified point.
     */
    public boolean containsPoint(Point point) {
        int x = getLocation().x + getSize().width / 2;
        int y = getLocation().y + getSize().height / 2;
        int r = Math.min(getSize().width / 2, getSize().height / 2);

        return (point.x - x) * (point.x - x) +
               (point.y - y) * (point.y - y) <= r * r;
    }

    /**
     * Calculates depth of the subtree with this node as root
     * assuming that this node has depth <code>curDepth</code>
     *
     * @param curDepth depth of this node.
     */
    public void calculateDepth(int curDepth) {
        depth = curDepth;

        NodeShape node = getLeftChildShape();

        while (node != null) {
            node.calculateDepth(depth + 1);
            node = node.getRightSiblingShape();
        }
    }

    public void calculateHorizontalPos() {
        hpos = tree.addElementAtDepth(depth);

        NodeShape node = getLeftChildShape();

        while (node != null) {
            node.calculateHorizontalPos();
            node = node.getRightSiblingShape();
        }
    }
}
