package ru.ifmo.vizi.ukkonen.SuffixTree;

import ru.ifmo.vizi.base.widgets.ShapeStyle;
import ru.ifmo.vizi.ukkonen.UkkonenVisualizer;

import java.awt.*;
import java.awt.event.MouseMotionAdapter;
import java.awt.event.MouseEvent;
import java.util.Hashtable;
import java.util.Vector;

/**
 * Suffix tree visualizer.
 *
 * @author  Igor Ahmetov
 * @version	$Id$
 */
public class SuffixTree extends Panel {
	/**
	 * Suffix tree.
	 */
	protected SuffixTreeData tree;

    /**
     * Shape styles.
     */
    protected ShapeStyle[] styleSet;

    /**
     * Hashtable for storing nodes of the tree.
     */
    protected Hashtable nodes = new Hashtable();

    /**
     * Current algorithm state.
     */
    protected SuffixTreeState state;

    /**
     * Stores number of elements on different levels.
     */
    protected Vector elemCount = new Vector();

    /**
     * Edge lock.
     */
    protected boolean edgeLock = false;

    /**
     * Edge that has acquired the lock.
     */
    protected Edge lockedEdge;

    /**
     * Used for double-buffering.
     */
    protected Image buffer;

    /**
     * Stores edges component of this tree.
     */
    protected Vector edges = new Vector();

    /**
     * Stores mouse cursor coordinates.
     */
    protected Point mousePoint = new Point(0, 0);

    /**
     * Parent applet.
     */
    protected UkkonenVisualizer applet;

	/**
	 * Default constructor.
	 *
	 * @param string	string of suffix tree.
	 */
	public SuffixTree(UkkonenVisualizer applet,
                      String string, ShapeStyle[] styleSet) {
		super(new SuffixTreeLayout());

        this.applet = applet;
        this.styleSet = styleSet;
		initTree(string);

        addMouseMotionListener(new TreeMouseMotionAdapter());
	}

    /**
     * Returns current string.
     */
    public String getString() {
        return tree.getString();
    }

    /**
     * Sets new string.
     */
    public void setString(String string) {
        removeAll();
        initTree(string);
        repaint();
    }

    /**
     * Initializes tree.
     *
     * @param string suffix tree string.
     */
    private void initTree(String string) {
        tree = new SuffixTreeData(string);
        state = tree.getState();
    }

    /**
     * Performs next step.
     */
    public void nextStep() {
        if (Options.DEBUG) System.out.println("SuffixTree.nextStep()");
        tree.nextStep();
        state = tree.getState();
        if (Options.DEBUG) System.out.println("New state:\n" + state);
    }

    /**
     * Returns to previous step.
     */
    public void previousStep() {
        if (Options.DEBUG) System.out.println("SuffixTree.previousStep()");
        tree.previousStep();
        state = tree.getState();
        if (Options.DEBUG) System.out.println("New state:\n" + state);
    }

    /**
     * Returns current algorithm state.
     *
     * @return current algorithm state.
     */
    public SuffixTreeState getState() {
        return state;
    }

    /**
     * Returns algorithm state during next step.
     *
     * @return algorithm state during next step.
     */
    public SuffixTreeState getNextState() {
        return tree.getNextState();
    }

    /**
     * Returns current extension.
     *
     * @return current extension.
     */
    public int getExtension() {
        return state.getExtension();
    }

    /**
     * Returns current phase.
     *
     * @return current phase.
     */
    public int getPhase() {
        return state.getPhase();
    }

    /**
     * Returns string representation of this tree.
     *
     * @return string representation of this tree.
     */
    public String toString() {
        return "Suffix tree:\n" + tree.toString() +
                "Current state:\n" + state.toString();
    }

    /**
     * Returns <code>NodeShape</code> object for given node.
     *
     * @param node suffix tree node.
     * @return <code>NodeShape</code> component for <code>node</code>.
     */
    public NodeShape getNodeShape(SuffixTreeNode node) {
        if (node == null) return null;
        return (NodeShape) nodes.get(node);
    }

    /**
     * Updates tree components.
     */
    public void updateTree() {
        if (Options.DEBUG) System.out.println("SuffixTree.updateTree");
        removeAll();
        nodes.clear();

        edges.removeAllElements();

        edgeLock = false;
        lockedEdge = null;

        updateNode(state.getRoot());
        updateLinks(state.getRoot());

        elemCount.removeAllElements();
        if (state.getRoot() != null) {
            getNodeShape(state.getRoot()).calculateDepth(1);
            getNodeShape(state.getRoot()).calculateHorizontalPos();
        }

        validate();

        for (int i = 0; i < edges.size(); i++) {
            Edge edge = (Edge) edges.elementAt(i);
            if (edge.containsPoint(mousePoint.x, mousePoint.y))
                edge.acquireLock();
        }
        repaint();

        if (Options.DEBUG) System.out.println("Components updated");
    }

    /**
     * Updates subtree with give node as its root.
     *
     * @param node root of the subtree.
     */
    protected void updateNode(SuffixTreeNode node) {
        if (node == null) return;

        NodeShape shape;
        if (node.isLeaf()) {
            shape = new NodeShape(styleSet, this, node,
                    String.valueOf(node.getLeafNumber()));
        } else {
            shape = new NodeShape(styleSet, this, node);
        }

        nodes.put(node, shape);

        int nodeStyle = state.getNodeStyle(shape.getNode());
        shape.setStyle(nodeStyle);
        add(shape);

        SuffixTreeNode child = node.getLeftChild();
        while (child != null) {
            updateNode(child);

            Edge edge = new Edge(styleSet, shape, getNodeShape(child));
            int edgeStyle = state.getEdgeStyle(edge.getTargetNodeShape().getNode());
            edge.setStyle(edgeStyle);

          	edge.setArrow(state.hasArrow(edge.getTargetNodeShape().getNode()));
          	edge.setReverseArrow(state.hasReverseArrow(edge.getTargetNodeShape().getNode()));

            add(edge);
            edges.addElement(edge);

            child = child.getRightSibling();
        }
    }

    /**
     * Updates links of the subtree with give node as its root.
     *
     * @param node root of the subtree.
     */
    protected void updateLinks(SuffixTreeNode node) {
        if (node == null) return;

        if (node.getSuffixLink() != null) {
            SuffixLink link = new SuffixLink(styleSet, getNodeShape(node),
                    getNodeShape(node.getSuffixLink()));

            int linkStyle = state.getLinkStyle(link.getSourceNodeShape().getNode());
            link.setStyle(linkStyle);
            add(link);
        }

        SuffixTreeNode child = node.getLeftChild();
        while (child != null) {
            updateLinks(child);
            child = child.getRightSibling();
        }
    }

    /**
     * Checks if the node corresponding to the specified <code>NodeShape</code>
     * object is selected in the current algorithm state.
     *
     * @param node target node.
     * @return <code>true</code> if <code>node</code> is selected,
     *         <code>false</code> otherwise.
     */
    public boolean isNodeSelected(NodeShape node) {
        return state.isNodeSelected(node.getNode());
    }

    /**
     * Adds new element at given depth.
     *
     * @param depth depth of the new element.
     * @return element position.
     */
    protected int addElementAtDepth(int depth) {
        if (elemCount.size() < depth) {
            elemCount.addElement(new Integer(1));
            return 1;
        } else {
            int pos = ((Integer) elemCount.elementAt(depth - 1)).intValue();
            elemCount.setElementAt(new Integer(pos + 1), depth - 1);
            return pos + 1;
        }
    }

    /**
     * Returns number of elements at given depth.
     * @param depth depth.
     * @return number of elements.
     */
    public int getElementCountAtDepth(int depth) {
        return ((Integer) elemCount.elementAt(depth - 1)).intValue();
    }

    /**
     * Tries to acquire edge lock.
     *
     * @param edge edge that wants to acquire lock.
     * @return <code>true</code> if lock has been succesfully acquired.
     */
    public boolean acquireLock(Edge edge) {
        if (!edgeLock) {
            edgeLock = true;
            lockedEdge = edge;
            return true;
        } else
            return false;
    }

    /**
     * Releases edge lock.
     *
     * @param edge edge that has acquired the lock.
     */
    public void releaseLock(Edge edge) {
        if (edge == lockedEdge) {
            lockedEdge = null;
            edgeLock = false;
        }
    }

    /**
     * Paints the suffix tree.
     *
     * @param g {@link Graphics} on which to paint.
     */
    public void paint(Graphics g) {
        int width = getSize().width;
        int height = getSize().height;

        if (width > 0 && height > 0) {
            if (buffer == null) {
                buffer = this.createImage(width, height);
            }

            Graphics bg = buffer.getGraphics();
            bg.clearRect(0, 0, width, height);

            super.paint(bg);
            if (lockedEdge != null)
                lockedEdge.paintHint(bg);

            bg.dispose();

            if (buffer != null && g != null)
                g.drawImage(buffer, 0, 0, null);
        }
    }

    /**
     * Invalidates this component.
     */
    public void invalidate() {
        super.invalidate();
        buffer = null;
    }

    /**
     * Updates the component.
     *
     * @param g {@link Graphics} on which to paint.
     */
    public void update(Graphics g) {
        paint(g);
    }

    /**
     * Returns mouse cursor point.
     *
     * @return mouse cursor point.
     */
    public Point getMousePoint() {
        return mousePoint;
    }

    /**
     * Moves specified component on top of this container.
     *
     * @param comp the component to be moved.
     */
    protected void moveToTop(Component comp) {
        remove(comp);
        add(comp, 0);
    }

    /**
     * Class for handling mouse motion events for the edge.
     */
    class TreeMouseMotionAdapter extends MouseMotionAdapter {
        /**
         * Handles mouse motion events.
         *
         * @param e <code>MouseEvent</code> object.
         */
        public void mouseMoved(MouseEvent e) {
            mousePoint = e.getPoint();
            for (int i = 0; i < edges.size(); i++) {
                Edge edge = (Edge) edges.elementAt(i);

                if (edge.containsPoint(e.getX(), e.getY())
                    && !edge.getSourceNodeShape().containsPoint(e.getPoint())
                    && !edge.getTargetNodeShape().containsPoint(e.getPoint())) {
                    if (edge.acquireLock()) {
                        SuffixTreeNode node = edge.getTargetNodeShape().getNode();
                        applet.stringSelect(node.getBeginIndex(),
                                            node.getEndIndex());
                    }
                    if (edge.hasLock()) {
                        moveToTop(edge);
                        moveToTop(edge.getSourceNodeShape());
                        moveToTop(edge.getTargetNodeShape());
                        repaint();
                    }
                } else {
                    if (edge.releaseLock()) {
                        applet.stringSelect(-1, -1);
                        repaint();
                    }
                }
            }
        }

        /**
         * Handles mouse dragging events.
         *
         * @param e <code>MouseEvent</code> object.
         */
        public void mouseDragged(MouseEvent e) {
            mousePoint = e.getPoint();
            for (int i = 0; i < edges.size(); i++) {
                Edge edge = (Edge) edges.elementAt(i);

                if (edge.containsPoint(e.getX(), e.getY())
                    && !edge.getSourceNodeShape().containsPoint(e.getPoint())
                    && !edge.getTargetNodeShape().containsPoint(e.getPoint())) {
                    if (edge.acquireLock()) {
                        SuffixTreeNode node = edge.getTargetNodeShape().getNode();
                        applet.stringSelect(node.getBeginIndex(),
                                            node.getEndIndex());
                    }
                    if (edge.hasLock()) {
                        moveToTop(edge);
                        moveToTop(edge.getSourceNodeShape());
                        moveToTop(edge.getTargetNodeShape());
                        repaint();
                    }
                } else {
                    if (edge.releaseLock()) {
                        applet.stringSelect(-1, -1);
                        repaint();
                    }
                }
            }
        }
    }
}