package ru.ifmo.vizi.ukkonen.SuffixTree;

import java.awt.*;
import java.util.Hashtable;

/**
 * Suffix tree layout.
 *
 * @author	Igor Ahmetov
 * @version	$Id$
 */
public class SuffixTreeLayout implements LayoutManager {
    /**
     * Horizontal gap.
     */
    protected double hgap = 0.06;

    /**
     * Vertical gap.
     */
    protected double vgap = 0.1;

    /**
     * Node size.
     */
    protected double nodeSize = 0.04;

    /**
     * Components that have been layed out.
     */
    protected Hashtable components = new Hashtable();

	/**
	 * Default constructor.
	 */
	public SuffixTreeLayout() {
	}

	/**
     * Adds the specified component to the layout.
     *
     * @param name the name of the component
     * @param comp the component to be added
     */
    public void addLayoutComponent(String name, Component comp) {
    }

    /**
     * Removes the specified component from the layout.
     *
     * @param comp the component to remove
     */
    public void removeLayoutComponent(Component comp) {
    }

    /**
     * Returns the preferred dimensions for this layout given the 
     * <i>visible</i> components in the specified target container.
     * @param target the component which needs to be laid out
     * @return    the preferred dimensions to lay out the
     *            subcomponents of the specified container
     */
    public Dimension preferredLayoutSize(Container target) {
    	synchronized (target.getTreeLock()) {
            return new Dimension(400, 400);
		}
    }

    /**
     * Returns the minimum dimensions needed to layout the <i>visible</i>
     * components contained in the specified target container.
     * @param target the component which needs to be laid out
     * @return    the minimum dimensions to lay out the
     *            subcomponents of the specified container
     */
    public Dimension minimumLayoutSize(Container target) {
	    synchronized (target.getTreeLock()) {
            return new Dimension(0, 0);
		}
    }

    /**
     * Layouts node <code>node</code> of tree <code>tree</code>, assuming
     * that depth of the tree is <code>depth</code>.
     *
     * @param tree suffix tree.
     * @param node node component to layout.
     * @param depth depth of the tree.
     * @param hoffset horizontal offset in the container.
     * @param width width of the container.
     * @param voffset vertical offset in the container.
     * @param height height of the container.
     */
    protected void layoutNode(SuffixTree tree, NodeShape node, int depth,
                             int hoffset, int width,
                             int voffset, int height) {
        if (node == null) return;

        double hpos = 0.5;
        double vpos = (double) (node.getDepth() - 1) / (depth - 1);

        if (Options.DEBUG) System.out.println("hpos: " + hpos + "; vpos: " + vpos);

        double size = Math.min(tree.getSize().width * nodeSize,
                tree.getSize().height * nodeSize);

        if (node.isVisible()) {
            if (node.getMessage().length() != 0 || tree.isNodeSelected(node)) {
                node.setBounds((int) (hoffset + width * hpos) - (int) (size * 1.25),
                        (int) (voffset + height * vpos) - (int) (size * 1.25),
                        (int) (size * 1.25) * 2 + 1, (int) (size * 1.25) * 2 + 1);
                node.adjustFontSize();
            } else {
                node.setBounds((int) (hoffset + width * hpos)- (int) size / 2,
                        (int) (voffset + height * vpos) - (int) size / 2,
                        (int) size / 2 * 2 + 1, (int) size / 2 * 2 + 1);
            }
            components.put(node, new Object());
        }

        int totalWidth = node.getNode().getWidth();
        int currentWidth = 0;

        NodeShape child = node.getLeftChildShape();

        while (child != null) {
            int nodeWidth = child.getNode().getWidth();
            layoutNode(tree, child, depth,
                    hoffset + (int) ((double) width * currentWidth / totalWidth),
                    (int) ((double) width * nodeWidth / totalWidth), voffset, height);

            currentWidth += nodeWidth;
            child = child.getRightSiblingShape();
        }
    }

    /**
     * Lays out the container. This method lets each component take
     * its preferred size by reshaping the components in the
     * target container in order to satisfy the alignment of
     * this <code>SuffixTreeLayout</code> object.
     *
     * @param target    the specified component being laid out
     */
    public void layoutContainer(Container target) {
    	synchronized (target.getTreeLock()) {
            if (Options.DEBUG) System.out.println("SuffixTreeLayout.layoutContainer()");
			Insets insets = target.getInsets();
            SuffixTree tree = (SuffixTree) target;
            Dimension size = target.getSize();

            components.clear();

			int width = size.width -
                    (insets.left + insets.right + (int) (size.width * hgap * 2));
            int height = size.height -
                    (insets.top + insets.bottom + (int) (size.height * vgap * 2));

			int nmembers = target.getComponentCount();

            if (Options.DEBUG) System.out.println("Laying out " + nmembers + " components");

            SuffixTreeNode root = tree.getState().getRoot();

            if (root != null) {
                int treeDepth = root.getDepth();

                layoutNode(tree, tree.getNodeShape(root), treeDepth,
                        insets.left + (int) (size.width * hgap), width,
                        insets.top + (int) (size.height * vgap), height);
            }

            for (int i = 0 ; i < nmembers ; i++) {
			    Component m = target.getComponent(i);
	    		if (m.isVisible() && !components.containsKey(m)) {
                    m.setBounds(0, 0, size.width, size.height);
			    }
			}
    	}
    }
}