package ru.ifmo.vizi.ukkonen.SuffixTree;

/**
 * Represents suffix tree node.
 *
 * @author		Igor Ahmetov
 * @version	$Id$
 */
public class SuffixTreeNode {
    /**
     * Parent node of this node.
     */
    protected SuffixTreeNode parent;

	/**
	 * Leftmost child of this node.
	 */
	protected SuffixTreeNode leftChild;

	/**
	 * Right sibling of this node.
	 */
	protected SuffixTreeNode rightSibling;

    /**
     * Suffix link.
     */
    protected SuffixTreeNode suffixLink;

	/**
	 * Start position of this node's substring.
	 */
	protected int beginIndex;

	/**
	 * End position of this node's substring.
	 */
	protected int endIndex;

	/**
	 * Suffix tree to which this node belongs.
	 */
	protected SuffixTreeData tree;

    /**
     * Leaf number of this node.
     */
    protected int leafNumber;

	/**
	 * Creates an inner node of a suffix tree..
	 *
	 * @param tree  suffix tree to which this node belongs
     * @param parent    parent node of this node.
	 * @param leftChild	leftmost child of this node.
	 * @param rightSibling	right sibling of this node.
     * @param suffixLink    suffix link of this node.
	 * @param beginIndex	start position of substring for this node.
	 * @param endIndex	end position of substring for this node.
	 */
	public SuffixTreeNode(SuffixTreeData tree,
                          SuffixTreeNode parent, SuffixTreeNode leftChild,
                          SuffixTreeNode rightSibling, SuffixTreeNode suffixLink,
                          int beginIndex, int endIndex) {
		this.beginIndex = beginIndex;
		this.endIndex = endIndex;

		this.parent = parent;
        this.leftChild = leftChild;
		this.rightSibling = rightSibling;
        this.suffixLink = suffixLink;

		this.tree = tree;
	}

    /**
	 * Constructs leaf node of a suffix tree.
	 *
	 * @param tree  suffix tree to which this node belongs
     * @param parent    parent node of this node.
     * @param rightSibling  right sibling of this node.
     * @param leafNumber    leaf number of this node.
	 * @param beginIndex	start position of substring for this node.
	 * @param endIndex	end position of substring for this node.
	 */
	public SuffixTreeNode(SuffixTreeData tree,
                          SuffixTreeNode parent, SuffixTreeNode rightSibling,
                          int leafNumber,
                          int beginIndex, int endIndex) {
		this.beginIndex = beginIndex;
		this.endIndex = endIndex;

		this.parent = parent;
        this.leafNumber = leafNumber;

        this.rightSibling = rightSibling;

		this.tree = tree;
	}

    /**
     * Returnes leaf number of this node if this node is a leaf.
     *
     * @return leaf number of this node.
     */
    public int getLeafNumber() {
        return leafNumber;
    }

    /**
     * Returns begin index of this node.
     *
     * @return begin index of this node.
     */
    public int getBeginIndex() {
    	return beginIndex;
    }

    /**
     * Sets begin index of this node.
     *
     * @param beginIndex new begin index of this node.
     */
    public void setBeginIndex(int beginIndex) {
        this.beginIndex = beginIndex;
    }

    /**
     * Returns end index node of this node.
     *
     * @return end index of this node.
     */
    public int getEndIndex() {
    	return endIndex;
    }

    /**
     * Sets end index of this node.
     *
     * @param endIndex new parent for this node.
     */
    public void setParent(int endIndex) {
        this.endIndex = endIndex;
    }

    /**
     * Returns parent node of this node.
     *
     * @return parent node of this node.
     */
    public SuffixTreeNode getParent() {
    	return parent;
    }

    /**
     * Sets parent of this node.
     *
     * @param parent new parent for this node.
     */
    public void setParent(SuffixTreeNode parent) {
        this.parent = parent;
    }

    /**
     * Returns leftmost child of this node.
     *
     * @return leftmost child of this node or <code>null</code> if none.
     */
    public SuffixTreeNode getLeftChild() {
    	return leftChild;
    }

    /**
     * Sets leftmost child of this node.
     *
     * @param leftChild new leftmost child of this node.
     */
    public void setLeftChild(SuffixTreeNode leftChild) {
        this.leftChild = leftChild;
    }

    /**
     * Returns right sibling of this node.
     *
     * @return right sibling of this node or <code>null</code> if none.
     */
    public SuffixTreeNode getRightSibling() {
    	return rightSibling;
    }

    /**
     * Sets right sibling of this node.
     *
     * @param rightSibling new right sibling of this node.
     */
    public void setRightSibling(SuffixTreeNode rightSibling) {
    	this.rightSibling = rightSibling;
    }

    /**
     * Returns suffix link of this node.
     *
     * @return suffix link of this node or <code>null</code> if none.
     */
    public SuffixTreeNode getSuffixLink() {
        return suffixLink;
    }

    /**
     * Sets suffix link of this node.
     *
     * @param suffixLink new suffix link of this node.
     */
    public void setSuffixLink(SuffixTreeNode suffixLink) {
        this.suffixLink = suffixLink;
    }

    /**
     * Returns left sibling of this node.
     *
     * @return left sibling of this node or <code>null</code> if none.
     */
    public SuffixTreeNode getLeftSibling() {
        SuffixTreeNode node = parent.getLeftChild();
        if (node == this)
            return null;
        while (node.getRightSibling() != this) {
            node = node.getRightSibling();
        }
    	return node;
    }

    /**
     * Returns rightmost child of this node.
     *
     * @return rightmost child of this node or <code>null</code> if none.
     */
    public SuffixTreeNode getRightChild() {
    	SuffixTreeNode node = getLeftChild();
        if (node == null)
            return null;

        while (node.getRightSibling() != null) {
            node = node.getRightSibling();
        }
        return node;
    }

    /**
     * Assures that there is an extension from this node that
     * starts woth given character and creates new node if there isn't one.
     *
     * @param index index of the character.
     */
    public void addChar(int index) {
        SuffixTreeNode node = getChild(index);
        if (node != null) return;

        if (leftChild == null) {
            if (parent == null) {
                leftChild = new SuffixTreeNode(tree, this, null, tree.getExtension() + 1,
                    index, index + 1);
            } else {
                endIndex++;
            }
        } else {
            leftChild = new SuffixTreeNode(tree, this, leftChild, tree.getExtension() + 1,
                        index, index + 1);
        }
    }


    /**
     * Returns depth of the subtree with root as this node.
     *
     * @return depth of the subtree with root as this node.
     */
    public int getDepth() {
        int depth = 0;
        SuffixTreeNode node = leftChild;

        while (node != null) {
            depth = Math.max(depth, node.getDepth());
            node = node.getRightSibling();
        }

        return depth + 1;
    }

    /**
     * Returns width of the subtree with root as this node.
     *
     * @return width of the subtree with root as this node.
     */
    public int getWidth() {
        int width = 0;
        SuffixTreeNode node = leftChild;

        while (node != null) {
            width = width + node.getWidth();
            node = node.getRightSibling();
        }

        if (width == 0)
            return 1;
        else
            return width;
    }

    /**
     * Returns substring on this node.
     *
     * @return	substring on this node.
     */
    public String getString() {
        if (parent == null)
            return "root";
        else
    	    return tree.getString().substring(beginIndex, endIndex);
    }

    /**
     * Returns length of the label on this node.
     *
     * @return length of the label on this node.
     */
    public int getLength() {
        return endIndex - beginIndex;
    }

    /**
     * Returns tree to which this node belongs.
     *
     * @return tree to which this node belongs.
     */
    public SuffixTreeData getTree() {
        return tree;
    }

    /**
     * Returns length of the label between root node and this node.
     *
     * @return length of the label between root node and this node.
     */
    public int getLabelLength() {
        if (parent == null)
            return getLength();
        else
            return getLength() + parent.getLabelLength();
    }

    /**
     * Returns <code>true</code> if this node is root node of the tree.
     *
     * @return <code>true</code> if this node is root node of the tree.
     */
    public boolean isRoot() {
        return parent == null;
    }

    /**
     * Returns <code>true</code> if this node is a leaf.
     *
     * @return <code>true</code> if this node is a leaf.
     */
    public boolean isLeaf() {
        return leftChild == null;
    }

    /**
     * Returns first char of this node.
     *
     * @return first char of this node.
     */
    public char getFirstChar() {
        return tree.getString().charAt(beginIndex);
    }

    /**
     * Returns child node that starts with given character.
     *
     * @param index index of the character.
     * @return child node or <code>null</code> if none.
     */
    public SuffixTreeNode getChild(int index) {
        SuffixTreeNode node = leftChild;
        while (node != null) {
            if (tree.getString().charAt(index) == node.getFirstChar()) {
                return node;
            }
            node = node.getRightSibling();
        }
        return null;
    }

    /**
     * Returns child node that starts with given character.
     *
     * @param ch character.
     * @return child node or <code>null</code> if none.
     */
    public SuffixTreeNode getChild(char ch) {
        SuffixTreeNode node = leftChild;
        while (node != null) {
            if (ch == node.getFirstChar()) {
                return node;
            }
            node = node.getRightSibling();
        }
        return null;
    }

    /**
     * Returns string representation of this node.
     *
     * @return string representation of this node.
     */
    public String toString() {
        String string = "Node: ";
        if (parent == null)
            string += "root; ";
        else
            string += getString() + "; parent: " + parent.getString() + "; ";

        if (suffixLink != null)
            string += "link: " + suffixLink.getString() + ";";
        string += "\nChildren: ";

        SuffixTreeNode node = leftChild;
        while (node != null) {
            string += node.getString() + " ";
            node = node.getRightSibling();
        }
        string += "\n\n";

        node = leftChild;
        while (node != null) {
            string += node.toString();
            node = node.getRightSibling();
        }

        return string;
    }

    /**
     * Checks if this node has suffix link.
     *
     * @return <code>true</code> if this node has suffix link.
     */
    public boolean hasSuffixLink() {
        return suffixLink != null;
    }
}