package ru.ifmo.vizi.ukkonen.SuffixTree;

import java.util.Hashtable;
import java.util.Vector;

/**
 * Represents current state of the Ukkonen algorithm.
 *
 * @author	Igor Ahmetov
 * @version	$Id$
 */
public class SuffixTreeState {
    /**
     * Style of leaf nodes.
     */
    public final static int LEAF_STYLE = 0;

    /**
     * Style of selected leaf nodes.
     */
    public final static int SELECTED_LEAF_STYLE = 1;

    /**
     * Style of inner nodes.
     */
    public final static int NODE_STYLE = 2;

    /**
     * Style of selected inner nodes.
     */
    public final static int SELECTED_NODE_STYLE = 3;

    /**
     * Style of edges.
     */
    public final static int EDGE_STYLE = 4;

    /**
     * Style of selected edges.
     */
    public final static int SELECTED_EDGE_STYLE = 5;

    /**
     * Style of suffix links.
     */
    public final static int LINK_STYLE = 6;

    /**
     * Style of selected suffix links.
     */
    public final static int SELECTED_LINK_STYLE = 7;

    /**
     * Default state.
     */
    public final static int STATE_NONE = 0;

    /**
     * Performing descent in the tree.
     */
    public final static int STATE_DESCENT = 1;

    /**
     * A char was added to the leaf label.
     */
    public final static int STATE_ADD_CHAR_LEAF = 2;

    /**
     * Edge was splitted.
     */
    public final static int STATE_ADD_CHAR_SPLIT_EDGE = 3;

    /**
     * New leaf was created during adding of a char to the tree.
     */
    public final static int STATE_ADD_CHAR_NEW_LEAF = 4;

    /**
     * Suffix link has been added to the tree.
     */
    public final static int STATE_LINK_ADDED = 5;

    /**
     * Phase of the algorithm.
     */
    protected int phase;

    /**
     * Extension number.
     */
    protected int extension;

    /**
     * State of the algorithm.
     */
    protected int state;

    /**
     * Root node.
     */
    protected SuffixTreeNode root;

    /**
     * Hashtable for matching nodes.
     */
    protected Hashtable map;

    /**
     * Selected suffix links.
     */
    protected Vector selectedLinks = new Vector();

    /**
     * Selected edges.
     */
    protected Vector selectedEdges = new Vector();

    /**
     * Selected nodes.
     */
    protected Vector selectedNodes = new Vector();

    /**
     * Edges with arrows.
     */
    protected Vector arrowEdges = new Vector();

    protected Vector reverseArrowEdges = new Vector();

    /**
     * Current suffix tree node.
     */
    protected SuffixTreeNode currentNode;

    /**
     * Active substring.
     */
    protected String activeSubstring = "";

    /**
     * Default constructor.
     *
     * @param phase phase of the algorithm.
     * @param extension extension number.
     * @param root root node of the tree.
     * @param currentNode current suffix tree node.
     */
    public SuffixTreeState(int phase, int extension, SuffixTreeNode root,
                           SuffixTreeNode currentNode) {
        this.phase = phase;
        this.extension = extension;

        map = new Hashtable();

        this.root = copyTree(root);
        copyLinks(root);

        if (currentNode != null)
            this.currentNode = (SuffixTreeNode) map.get(currentNode);

        if (Options.DEBUG) {
            System.out.println("SuffixTreeState()");
            System.out.println("Stored state:\n" + this.root);
        }
    }

    /**
     * Adds given node to the list of selected nodes.
     *
     * @param node target node.
     */
    public void selectNode(SuffixTreeNode node) {
        selectedNodes.addElement(map.get(node));
    }

    /**
     * Adds the suffix link from a given node to the list of selected links.
     *
     * @param node source of the link.
     */
    public void selectLink(SuffixTreeNode node) {
        selectedLinks.addElement(map.get(node));
    }

    /**
     * Adds edge from the given node to its parent to the list of selected edges.
     *
     * @param node end of the edge.
     */
    public void selectEdge(SuffixTreeNode node) {
        selectedEdges.addElement(map.get(node));
    }

    /**
     * Adds arrow to the edge.
     *
     * @param edge.
     */
    public void addArrow(SuffixTreeNode node) {
        arrowEdges.addElement(map.get(node));
    }

    public void addReverseArrow(SuffixTreeNode node) {
    	reverseArrowEdges.addElement(map.get(node));
    }

    /**
     * Copies suffix tree without suffix links.
     *
     * @param node root of the tree to copy.
     * @return root of the new tree.
     */
    protected SuffixTreeNode copyTree(SuffixTreeNode node) {
        if (node == null) return null;

        SuffixTreeNode newNode;

        if (node.getParent() == null) {
            if (node.isLeaf()) {
                newNode = new SuffixTreeNode(node.getTree(),
                        null, copyTree(node.getRightSibling()), node.getLeafNumber(),
                        node.getBeginIndex(), node.getEndIndex());
            } else {
                newNode = new SuffixTreeNode(node.getTree(),
                        null, null, copyTree(node.getRightSibling()), null,
                        node.getBeginIndex(), node.getEndIndex());
            }
        } else {
            if (node.isLeaf()) {
                newNode = new SuffixTreeNode(node.getTree(),
                        (SuffixTreeNode) map.get(node.getParent()),
                        copyTree(node.getRightSibling()), node.getLeafNumber(),
                        node.getBeginIndex(), node.getEndIndex());
            } else {
                newNode = new SuffixTreeNode(node.getTree(),
                        (SuffixTreeNode) map.get(node.getParent()),
                        null, copyTree(node.getRightSibling()), null,
                        node.getBeginIndex(), node.getEndIndex());
            }
        }

        map.put(node, newNode);
        newNode.setLeftChild(copyTree(node.getLeftChild()));

        return newNode;
    }

    /**
     * Corrects links.
     *
     * @param node root of the old tree.
     */
    protected void copyLinks(SuffixTreeNode node) {
        if (node == null) return;

        SuffixTreeNode newNode = (SuffixTreeNode) map.get(node);
        if (node.getSuffixLink() != null)
            newNode.setSuffixLink((SuffixTreeNode) map.get(node.getSuffixLink()));

        copyLinks(node.getLeftChild());
        copyLinks(node.getRightSibling());
    }

    /**
     * Returns current algorithm phase.
     *
     * @return algorithm phase.
     */
    public int getPhase() {
        return phase;
    }

    /**
     * Returns current extension number.
     *
     * @return current extension number.
     */
    public int getExtension() {
        return extension;
    }

    /**
     * Returns current algorithm special state.
     *
     * @return algorithm state.
     */
    public int getSpecialState() {
        return state;
    }

    /**
     * Sets current algorithm special state.
     *
     * @param state new state.
     */
    public void setSpecialState(int state) {
        this.state = state;
    }

    /**
     * Returns suffix tree root.
     *
     * @return suffix tree root.
     */
    public SuffixTreeNode getRoot() {
        return root;
    }

    /**
     * Retuns current node.
     *
     * @return current node.
     */
    public SuffixTreeNode getCurrentNode() {
        return currentNode;
    }

    /**
     * Returns string representation of this <code>SuffixTreeState</code> object
     *
     * @return string representation of this object.
     */
    public String toString() {
        if (root != null)
            return root.toString();
        else
            return "null\n";
    }

    /**
     * Returns style of the node <code>node</code>.
     *
     * @param node target.
     * @return style of the node.
     */
    public int getNodeStyle(SuffixTreeNode node) {
        if (node.isLeaf()) {
            if (selectedNodes.contains(node))
                return SELECTED_LEAF_STYLE;
            else
                return LEAF_STYLE;
        } else {
            if (selectedNodes.contains(node))
                return SELECTED_NODE_STYLE;
            else
                return NODE_STYLE;
        }
    }

    /**
     * Returns style of the suffix link from node <code>node</code>.
     *
     * @param node target.
     * @return style of the link from specified node.
     */
    public int getLinkStyle(SuffixTreeNode node) {
        if (selectedLinks.contains(node))
            return SELECTED_LINK_STYLE;
        else
            return LINK_STYLE;
    }

    /**
     * Returns style of the edge between node <code>node</code> and its parent.
     *
     * @param node target.
     * @return style of the edge between specified node and its parent.
     */
    public int getEdgeStyle(SuffixTreeNode node) {
        if (selectedEdges.contains(node))
            return SELECTED_EDGE_STYLE;
        else
            return EDGE_STYLE;
    }

    /**
     * Returns <code>true</code> if given node is selected.
     *
     * @param node node to check for selection.
     * @return <code>true</code> if specified node is selected.
     */
    public boolean isNodeSelected(SuffixTreeNode node) {
        return selectedNodes.contains(node);
    }

    /**
     * Returns <code>true</code> if given edge contains arrow.
     *
     * @param node check edge from this node to its parent.
     * @return <code>true</code> if specified edge contains arrow.
     */
    public boolean hasArrow(SuffixTreeNode node) {
        return arrowEdges.contains(node);
    }

    public boolean hasReverseArrow(SuffixTreeNode node) {
    	return reverseArrowEdges.contains(node);
    }

    /**
     * Sets active substring.
     *
     * @param substring new active substring.
     */
    public void setActiveSubstring(String substring) {
        activeSubstring = substring;
    }

    /**
     * Returns active substring of this state.
     *
     * @return active substring.
     */
    public String getActiveSubstring() {
        return activeSubstring;
    }
}