package ru.ifmo.vizi.ukkonen;

import ru.ifmo.vizi.base.ui.*;
import ru.ifmo.vizi.base.*;
import ru.ifmo.vizi.base.widgets.ShapeStyle;
import ru.ifmo.vizi.ukkonen.widgets.StringVisualizer;
import ru.ifmo.vizi.ukkonen.widgets.VerticalStringVisualizer;
import ru.ifmo.vizi.ukkonen.ui.HintedChoice;
import ru.ifmo.vizi.ukkonen.ui.HintedTextField;
import ru.ifmo.vizi.ukkonen.ui.HintedCheckbox;
import ru.ifmo.vizi.ukkonen.SuffixTree.SuffixTree;
import ru.ifmo.vizi.ukkonen.SuffixTree.SuffixTreeState;

import java.awt.*;
import java.awt.event.*;
import java.util.Vector;

/**
 * Ukkonen applet.
 *
 * @author  Igor Ahmetov
 * @version $Id$
 */
public final class UkkonenVisualizer extends Base {
    /**
     * Suffix tree sentinel.
     */
    public static final String SENTINEL = "$";

    /**
     * Ukkonen automata instance.
     */
    private final Ukkonen auto;

    /**
     * Ukkonen automata data.
     */
    private final Ukkonen.Data data;

    /**
     * Save/load dialog.
     */
    private SaveLoadDialog saveLoadDialog;

    /**
     * Predefined examples.
     */
    private Vector examples = new Vector();

    /**
     * String visualizator.
     */
    private StringVisualizer stringVis;

    /**
     * Active string visualizer.
     */
    protected VerticalStringVisualizer activeStringVis;


    /**
     * Suffix tree.
     */
    private SuffixTree suffixTree;

    /**
     *
     */
    protected ClientAdapter clientAdapter;

    /**
     * Boolean variable that is <code>true</code> if the string field has focus.
     */
    protected boolean stringFieldHasFocus = false;

    /**
     * Component with focus.
     */
    protected Component focusComponent;


    /**
     * Creates a new Ukkonen visualizer.
     *
     * @param parameters visualizer parameters.
     */
    public UkkonenVisualizer(VisualizerParameters parameters) {
        super(parameters);
        auto = new Ukkonen(locale);
        data = auto.d;
        data.visualizer = this;

        clientPane.addComponentListener(
                clientAdapter = new ClientAdapter(config.getDouble("string-visualizer-height", 0.2),
                                  config.getDouble("string-visualizer-width", 0.1))
        );

        stringVis = new StringVisualizer(
                this, ShapeStyle.loadStyleSet(config, "string-visualizer")
        );
        activeStringVis = new VerticalStringVisualizer(
                this, ShapeStyle.loadStyleSet(config, "string-visualizer"), false
        );
        suffixTree = new SuffixTree(this, "", ShapeStyle.loadStyleSet(config, "suffix-tree"));

        clientPane.setLayout(new BorderLayout());
        clientPane.add(suffixTree, BorderLayout.CENTER);
        clientPane.add(activeStringVis, BorderLayout.EAST);
        clientPane.add(stringVis, BorderLayout.SOUTH);

        createInterface(auto);
    }

    /**
     * Performs next step.
     */
    public void doNextStep() {
        if (!stringFieldHasFocus) {
            super.doNextStep();
        }
    }

    /**
     * Performs next big step.
     */
    public void doNextBigStep() {
	    if (!stringFieldHasFocus) {
            super.doNextBigStep();
        }
    }

    /**
     * Performs previous step.
     */
    public void doPrevStep() {
        if (!stringFieldHasFocus) {
            super.doPrevStep();
        }
    }

    /**
     * Performs previous big step.
     */
    public void doPrevBigStep() {
    	if (!stringFieldHasFocus) {
    		super.doPrevBigStep();
    	}
    }

    /**
     * This method creates panel with visualizer controls.
     *
     * @return controls pane.
     */
    public Component createControlsPane() {
        Panel panel = new Panel(new BorderLayout());
        Panel bottomPanel = new Panel(new FlowLayout());

        panel.add(new AutoControlsPane(config, auto, forefather, true), BorderLayout.CENTER);

        HintedCheckbox stringCheckbox = new HintedCheckbox(config, "string-checkbox");
        stringCheckbox.setState(true);
        stringCheckbox.addItemListener(
                new ItemListener() {
                    public void itemStateChanged(ItemEvent e) {
                        stringVis.setVisible(e.getStateChange() == ItemEvent.SELECTED);
                        clientPane.validate();
                        updateTree();
                    }
                }
        );
        bottomPanel.add(stringCheckbox);

        HintedTextField stringField = new HintedTextField(config, "string-field", "", 24);
        stringField.addActionListener(
                new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                        setString(((HintedTextField) e.getSource()).getText());
                    }
                }
        );
        stringField.addKeyListener(
                new KeyAdapter() {
                    public void keyPressed(KeyEvent e) {
                        char ch = e.getKeyChar();
                        if (!(Character.isLetter(ch)
                                || Character.getType(ch) == Character.CONTROL
                                || Character.getType(ch) == Character.FORMAT
                                || Character.getType(ch) == Character.MODIFIER_LETTER
                                || Character.getType(ch) == Character.MODIFIER_SYMBOL
                                || Character.getType(ch) == Character.UNASSIGNED))
                            e.consume();

                    }
                }
        );
        stringField.addFocusListener(
                new FocusListener() {
                    public void focusGained(FocusEvent e) {
                        stringFieldHasFocus = true;
                    }

                    public void focusLost(FocusEvent e) {
                        stringFieldHasFocus = false;
                    }
                }
        );
        bottomPanel.add(stringField);

        final HintedChoice examplesChoice = new HintedChoice(config, "examples");
        examplesChoice.addItemListener(new ItemListener() {
            public void itemStateChanged(ItemEvent e) {
                setString((String) examples.elementAt(examplesChoice.getSelectedIndex()));
            }
        });

        int num = config.getInteger("examples-num", 0);
        for (int i = 0; i < num; i++) {
            examplesChoice.add(config.getParameter("example-" + (i + 1) + "-title", "Example " + i));
            examples.addElement(config.getParameter("example-" + (i + 1), ""));
        }
        setString((String) examples.elementAt(0));

        bottomPanel.add(examplesChoice);

        if (config.getBoolean("button-ShowSaveLoad")) {
            bottomPanel.add(new HintedButton(config, "button-SaveLoad") {
                protected void click() {
                    saveLoadDialog.center();
                    StringBuffer buffer = new StringBuffer();
                    buffer.append("/* ").append(
                            config.getParameter("save-string-comment")
                    ).append(" */\n");

                    String str = getString();
                    buffer.append("string = " + str.substring(0, str.length() - 1));

                    buffer.append("\n/* ").append(
                            config.getParameter("save-step-comment")
                    ).append(" */\n");
                    buffer.append("step = ").append(auto.getStep());
                    saveLoadDialog.show(buffer.toString());
                }
            });
        }
        panel.add(bottomPanel, BorderLayout.SOUTH);

        saveLoadDialog = new SaveLoadDialog(config, forefather) {
            public boolean load(String text) throws Exception {
                SmartTokenizer tokenizer = new SmartTokenizer(text, config);
                tokenizer.expect("string");
                tokenizer.expect("=");
                setString(tokenizer.nextWord());

                tokenizer.expect("step");
                tokenizer.expect("=");
                rewind(tokenizer.nextInt());

                tokenizer.expectEOF();

                return true;
            }
        };

        return panel;
    }

    /**
     * Sets new string for the suffix tree.
     *
     * @param string new string.
     */
    public void setString(String string) {
        rewind(0);
        suffixTree.setString(string + SENTINEL);
        stringVis.setString(string);
        //activeStringVis.setVisible(false);
    }

    /**
     * Rewinds algorithm to the specified step.
     *
     * @param step spet of the algorith to rewind to.
     */
    private void rewind(int step) {
        auto.toStart();

        while (!auto.isAtEnd() && auto.getStep() < step) {
            auto.stepForward(0);
        }
    }

    /**
     * Returns suffix tree string.
     *
     * @return suffix tree string.
     */
    public String getString() {
        return suffixTree.getString();
    }

    /**
     * Performs next step.
     */
    public void nextStep() {
        suffixTree.nextStep();
    }

    /**
     * Returns to previous step.
     */
    public void previousStep() {
        suffixTree.previousStep();
    }

    /**
     * Updates string visualizers.
     */
    protected void updateStrings() {
        stringVis.highlight(suffixTree.getExtension(), suffixTree.getPhase() + 2);

        SuffixTreeState state = suffixTree.getState();
        String str = state.getActiveSubstring();
        if (state.getSpecialState() == SuffixTreeState.STATE_DESCENT) {
            str = state.getCurrentNode().getString() + str;
        }

        activeStringVis.setString(str, false);
        if (state.getSpecialState() == SuffixTreeState.STATE_DESCENT) {
        	activeStringVis.highlight(0, state.getCurrentNode().getString().length());
        }
        /*if (str.length() == 0) {
            if (activeStringVis.isVisible()) {
                activeStringVis.setVisible(false);
                clientPane.validate();
            }
        } else {
            boolean visible = activeStringVis.isVisible();
            activeStringVis.setVisible(true);
            
            if (!visible) clientPane.validate();
        }*/
    }
    /**
     * Returns suffix tree.
     *
     * @return current suffix tree.
     */
    public SuffixTree getSuffixTree() {
        return suffixTree;
    }

    /**
     * Selects part of the string.
     *
     * @param beginIndex begin index.
     * @param endIndex end index.
     */
    public void stringSelect(int beginIndex, int endIndex) {
        stringVis.select(beginIndex, endIndex);
    }

    /**
     * Updates the suffix tree.
     */
    public void updateTree() {
        suffixTree.updateTree();
        updateStrings();
    }

    /**
     * Adds sentinel character to string visualizer string.
     */
    public void addSentinelCharacter() {
        stringVis.setString(suffixTree.getString());
    }

    /**
     * Returns current algorithm state.
     *
     * @return current algorithm state.
     */
    public SuffixTreeState getState() {
        return suffixTree.getState();
    }

    /**
     * Returns next algorithm state.
     * @return next algorithm state.
     */
    public SuffixTreeState getNextState() {
        return suffixTree.getNextState();
    }

    /**
     * Removes sentinel character from string visualizer string.
     */
    public void removeSentinelCharacter() {
        String str = suffixTree.getString();
        stringVis.setString(str.substring(0, str.length() - 1));
    }

    /**
     * Validates this container.
     */
    public void resizeStringVisualizers() {
	    if (clientPane.getSize().height != 0) {
    		clientAdapter.setStringVisualizerHeight(
    			(double) stringVis.getSize().height / clientPane.getSize().height
    		);
    	}
    	if (clientPane.getSize().width != 0) {
    		clientAdapter.setStringVisualizerWidth(
    			(double) activeStringVis.getSize().width / clientPane.getSize().width
    		);
    	}
    }

    /**
     * Returns some information about the applet.
     *
     * @return string containing short description of the applet.
     */
    public String getAppletInfo() {
        return "Visualizer of the Ukkonen Algorithm. Igor Ahmetov, CTD IFMO, 2004";
    }

    
    /**
     * Component adapter that handles resizing of the client pane.
     */
    class ClientAdapter extends ComponentAdapter {
        /**
         * Height of the string visualizer.
         */
        protected double stringVisHeight;

        /**
         * Width of the vertical string visualizer.
         */
        protected double stringVisWidth;

        /**
         * Default constructor.
         *
         * @param height height of the string visualizer
         */
        public ClientAdapter(double height, double width) {
            this.stringVisHeight = height;
            this.stringVisWidth = width;
        }        

        /**
         * Sets height of the horizontal string visualizer.
         *
         * @param height new height of the visualizer.
         */
        public void setStringVisualizerHeight(double height) {
        	stringVisHeight = height;
        }

        /**
         * Sets width of the vertical string visualizer.
         *
         * @param width new width of the visualizer.
         */
        public void setStringVisualizerWidth(double width) {
        	stringVisWidth = width;
        }

        /**
         * Deals with resizing of the clientPane.
         * @param e component event
         */
        public void componentResized(ComponentEvent e) {
            stringVis.setSize(200,
                    (int) (e.getComponent().getSize().height * stringVisHeight));
            stringVis.resize();
            activeStringVis.setSize(
                    (int) (e.getComponent().getSize().width * stringVisWidth), 200);
            activeStringVis.resize();
            e.getComponent().validate();
        }
    }
}