package ru.ifmo.vizi.ukkonen.widgets;

import ru.ifmo.vizi.base.widgets.Rect;
import ru.ifmo.vizi.base.widgets.ShapeStyle;
import ru.ifmo.vizi.ukkonen.UkkonenVisualizer;

import java.util.Vector;
import java.awt.*;
import java.awt.event.*;

/**
 * Class for string visualization.
 *
 * @author  Igor Ahmetov
 */

public class VerticalStringVisualizer extends Panel {
    /**
     * Panel style.
     */
    protected final static int VISUALIZER_STYLE = 0;
    /**
     * Default style for character cells.
     */
    protected final static int DEFAULT_STYLE = 1;

    /**
     * Selected style.
     */
    protected final static int HIGHLIGHT_STYLE = 2;

    /**
     * Highlight style for character cells.
     */
    protected final static int SELECT_STYLE = 3;

    /**
     * Numbers style.
     */
    protected final static int NUMBERS_STYLE = 4;

    /**
     * Width of the component, used for resizing with mouse.
     */
    protected int width;

    /**
     * New width of the component, used for resizing with mouse.
     */
    protected int newWidth;

    /**
     * Is <code>true</code> when the component is being resized.
     */
    protected boolean resizing = false;

    /**
     * Used for holding the elements.
     */
    protected Vector cells;

    /**
     * Used for holding components that represent numbers.
     */
    protected Vector numbers;

    /**
     * Current string.
     */
    protected String string;

    /**
     * Cells styleset.
     */
    protected ShapeStyle[] styleSet;

    /**
     * Used in double-buffering.
     */
    protected Image buffer;

    /**
     * Begin index of the highlighted part of the string.
     */
    protected int highlightBeginIndex;

    /**
     * End index of the highlighted part of the string.
     */
    protected int highlightEndIndex;

    /**
     * Begin index of the selected part of the string.
     */
    protected int selectBeginIndex;

    /**
     * End index of the selected part of the string.
      */
    protected int selectEndIndex;

    /**
     * Parent applet.
     */
    protected UkkonenVisualizer applet;

    /**
     * Default constructor.
     *
     * @param styleSet cells styleset.
     * @param showNumbers must be <code>true</code> if we need to show the numbers of the cells.
     */
    public VerticalStringVisualizer(UkkonenVisualizer applet, ShapeStyle[] styleSet, boolean showNumbers) {
        setLayout(null);

        this.styleSet = styleSet;
        this.applet = applet;

        this.addComponentListener(new ComponentAdapter() {
            public void componentResized(ComponentEvent e) {
                VerticalStringVisualizer.this.resize();
            }
        });

        setString("", showNumbers);

        addMouseMotionListener(new StringMouseMotionAdapter());
        addMouseListener(new StringMouseAdapter());
    }

    /**
     * Sets string.
     *
     * @param string new string.
     * @param showNumbers must be <code>true</code> if we need to show the numbers.
     */
    public void setString(String string, boolean showNumbers) {
    	this.string = string;
        removeAll();

        cells = new Vector();
        numbers = new Vector();
        for (int i = 0; i < string.length(); i++) {
        	Rect elem = new Rect(styleSet, string.substring(i, i + 1));
            elem.setStyle(DEFAULT_STYLE);
        	cells.addElement(elem);

            add(elem);

            if (showNumbers) {
                Rect number = new Rect(styleSet, String.valueOf(i + 1));
                number.setStyle(NUMBERS_STYLE);
                numbers.addElement(number);

                add(number);
            }
        }

        resize();
        repaint();
    }

    /**
     * Returnes current string.
     *
     * @return string current string.
     */
    public String getString() {
    	return string;
    }

    /**
     * Higlights cells from <code>beginIndex</code> to
     * <code>endIndex - 1</code> (inclusive).
     *
     * @param beginIndex
     * @param endIndex
     */
    public void highlight(int beginIndex, int endIndex) {
        highlightBeginIndex = beginIndex;
        highlightEndIndex = endIndex;

        for (int i = 0; i < cells.size(); i++) {
            ((Rect) cells.elementAt(i)).setStyle(DEFAULT_STYLE);
        }
        for (int i = Math.max(0, beginIndex); i < Math.min(endIndex, cells.size()); i++) {
            ((Rect) cells.elementAt(i)).setStyle(HIGHLIGHT_STYLE);
        }
        for (int i = Math.max(0, selectBeginIndex);
             i < Math.min(selectEndIndex, cells.size()); i++) {
            ((Rect) cells.elementAt(i)).setStyle(SELECT_STYLE);
        }

        repaint();
    }

    public void select(int beginIndex, int endIndex) {
        selectBeginIndex = beginIndex;
        selectEndIndex = endIndex;

        for (int i = 0; i < cells.size(); i++) {
            ((Rect) cells.elementAt(i)).setStyle(DEFAULT_STYLE);
        }
        for (int i = Math.max(0, highlightBeginIndex);
             i < Math.min(highlightEndIndex, cells.size()); i++) {
            ((Rect) cells.elementAt(i)).setStyle(HIGHLIGHT_STYLE);
        }
        for (int i = Math.max(0, beginIndex); i < Math.min(endIndex, cells.size()); i++) {
            ((Rect) cells.elementAt(i)).setStyle(SELECT_STYLE);
        }

        repaint();
    }

    /**
     * Resizes the component. Must be called every time after the
     * component has been resized.
     */
    public void resize() {
        int n = cells.size();
        Dimension size = getSize();

        int hskip = (int) Math.round((double) size.width / 8);
        int vskip = 10;

        int height, numWidth = 0;

        if (numbers.size() > 0) {
            height = (int) Math.min(Math.round((double) (size.height - 2 * vskip) / n),
                    Math.round((double) (size.width - 2 * hskip) * 5 / 8));
            numWidth =  (int) Math.round((double) 3 / 5 *  height);
        } else {
            height = (int) Math.min(Math.round((double) (size.height - 2 * vskip) / n),
                    Math.round((double) (size.width - 2 * hskip)));
        }

        int vgap = (size.height - height * n) / 2;

        for (int i = 0; i < n; i++) {
            Rect rect = (Rect) cells.elementAt(i);
            rect.setBounds(hskip, vgap + i * height, height + 1, height + 1);
            rect.adjustFontSize();

            if (numbers.size() > 0) {
                rect = (Rect) numbers.elementAt(i);
                rect.setBounds(hskip + width, vgap + i * height, height + 1, numWidth + 1);
                rect.adjustFontSize();
            }
        }
    }

    /**
     * Paints the string tree.
     *
     * @param g {@link java.awt.Graphics} on which to paint.
     */
    public void paint(Graphics g) {
        int width = getSize().width;
        int height = getSize().height;

        if (width > 0 && height > 0) {
            if (buffer == null) {
                buffer = this.createImage(width, height);
            }

            Graphics bg = buffer.getGraphics();

            if (styleSet[VISUALIZER_STYLE].getFillStatus()) {
                bg.setColor(styleSet[VISUALIZER_STYLE].getFillColor());
            }
            bg.fillRect(0, 0, width, height);

            super.paint(bg);

            if (styleSet[VISUALIZER_STYLE].getBorderStatus()) {
                bg.setColor(styleSet[VISUALIZER_STYLE].getBorderColor());
                bg.drawLine(0, 0, 0, height);
            }

            bg.dispose();

            g.drawImage(buffer, 0, 0, null);
        }
    }

    /**
     * Invalidates this component.
     */
    public void invalidate() {
        super.invalidate();
        buffer = null;
    }

    /**
     * Updates the component.
     *
     * @param g {@link Graphics} on which to paint.
     */
    public void update(Graphics g) {
        paint(g);
    }

    /**
     * Supporting class for handling mouse motion events for the string visualizer.
     *
     * @author  Igor Ahmetov
     * @version	$Id$
     */
    class StringMouseMotionAdapter extends MouseMotionAdapter {
        /**
         * Handles mouse dragging.
         *
         * @param e mouse event.
         */
        public void mouseDragged(MouseEvent e) {
            if (resizing) {
                if (width - e.getX() > 5 && width - e.getX() < getParent().getSize().width - 5)
                    newWidth = width - e.getX();
            }
        }

        /**
         * Handles mouse movement.
         *
         * @param e mouse event.
         */
        public void mouseMoved(MouseEvent e) {
            if (e.getX() < 10) {
                setCursor(new Cursor(Cursor.E_RESIZE_CURSOR));
            } else {
                setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
            }
        }
    }

    /**
     * Supporting class for handling mouse events for the string visualizer.
     *
     * @author  Igor Ahmetov
     * @version	$Id$
     */
    class StringMouseAdapter extends MouseAdapter {
        /**
         * Handles mouse clicks.
         *
         * @param e mouse event.
         */
        public void mousePressed(MouseEvent e) {
            if (e.getX() < 10) {
                resizing = true;
                newWidth = width = getSize().width;
            }
        }

        /**
         * Handles mouse clicks.
         * @param e mouse event.
         */
        public void mouseReleased(MouseEvent e) {
            if (resizing) {
                resizing = false;
                setSize(newWidth, getSize().height);
                applet.resizeStringVisualizers();
                getParent().validate();
            }
        }
    }
}